from .plotter import Plotter, StackedPlotter
from .PCA import PCA, PCA_Signalum
from .FFT import FFT, CWT
from .Basics import Basics
from .Calc import Calc
from .HTTP import HTTP

from .DataManipulation import DataManipulation
from .DataManipulation import Dataframe, Array


__version__ = "0.2.5"